# sage_setup: distribution = sagemath-graphs
r"""
Hypergraph isomorphic copy search

This module implements a code for the following problem:

    **INPUT:** two hypergraphs `H_1, H_2`

    **OUTPUT:** a copy of `H_2` in `H_1`

It is also possible to enumerate all such copies, and to require that such
copies be induced copies. More formally:

    A copy of `H_2` in `H_1` is an injection `f:V(H_2)\mapsto V(H_1)` such that
    for any set `S_2\in E(H_2)` we have `f(S_2)\in E(H_1)`.

    It is an *induced* copy if no other set of `E(H_1)` is contained in
    `f(V(H_2))`, i.e. `|E(H_2)|=\{S:S\in E(H_1)\text{ and }S\subseteq
    f(V(H_2))\}`.

The functions implemented here lists all such injections. In particular, the
number of copies of `H` in itself is equal to `|Aut(H)|`.

The feature is available through
:meth:`IncidenceStructure.isomorphic_substructures_iterator`.

Implementation
--------------

A hypergraph is stored as a list of edges, each of which is a "dense" bitset
over `|V(H_1)|` points. In particular, two sets of distinct cardinalities
require the same memory space. A hypergraph is a C struct with the following
fields:

* ``n``, ``m`` -- (``int``) number of points and edges

* ``limbs`` -- (``int``) number of 64-bits blocks per set

* ``set_space`` -- (``uint64_t *``) address of the memory used to store the
  sets

* ``sets`` -- (``uint64_t **``) ``sets[i]`` points toward the ``limbs``
  blocks encoding set `i`. Note also that ``sets[i][limbs]`` is equal to the
  cardinality of ``set[i]``, so that ``sets`` has length
  ``m*(limbs+1)*sizeof(uint64_t)``.

* ``names`` -- (``int *``) associates an integer 'name' to each of the ``n``
  points

The operations used on this data structure are:

* ``void permute(hypergraph * h, int n1, int n2)`` -- exchanges points `n1`
  and `n2` in the data structure. Note that their names are also exchanged
  so that we still know which is which.

* ``int induced_hypergraph(hypergraph * h1, int n, hypergraph * tmp1)`` --
  stores in ``tmp1`` the hypergraph induced by the first `n` points,
  i.e. all sets `S` such that `S\subseteq \{0,...,n-1\}`. The function
  returns the number of such sets.

* ``void trace_hypergraph64(hypergraph * h, int n, hypergraph * tmp)`` -- stores
  in ``tmp1`` the trace of `h` on the first `n` points, i.e. all sets of the
  form `S\cap \{0, \ldots, n-1\}`.

Algorithm
---------

We try all possible assignments of a representant `r_i\in H_1` for every `i\in
H_2`. When we have picked a representant for the first `n<` points
`\{0, \ldots, n-1\}\subsetneq V(H_2)`, we check that:

* The hypergraph induced by the (ordered) list `0, \ldots, n-1` in `H_2` is
  equal to the one induced by `r_0, \ldots, r_{n-1}` in `H_1`.

* If `S\subseteq \{0,...,n-1\}` is contained in `c` sets of size `k` in
  `H_2`, then `\{r_i:i\in S\}` is contained in `\geq c` sets of size `k` in
  `H_1`. This is done by comparing the trace of the hypergraphs while
  remembering the original size of each set.

As we very often need to build the hypergraph obtained by the trace of the first
`n` points (for all possible `n`), those hypergraphs are cached. The hypergraphs
induced by the same points are handled similarly.

Limitations
-----------

**Number of points** For efficiency reason the implementation assumes that `H_2`
has `\leq 64` points. Making this work for larger values means that calls to
``qsort`` have to be replaced by calls to ``qsort_r`` (i.e. to sort the edges
you need to know the number of limbs per edge) and that induces a big slowdown
for small cases (~50% when this code was implemented). Also, 64 points for `H_2`
is already very very big considering the problem at hand. Even `|V(H_1)|> 64`
seems too much.

**Vertex ordering** The order of vertices in `H_2` has a huge influence on the
performance of the algorithm. If no set of `H_2` contains more that one of the
first `k<n` points, then almost all partial assignments of representants are
possible for the first `k` points (though the degree of the vertices is taken
into account). For this reason it is best to pick an ordering such that the
first vertices are contained in as many sets as possible together. A heuristic
is implemented at
:meth:`~sage.combinat.designs.subhypergraph_search.SubHypergraphSearch.relabel_heuristic`.

AUTHORS:

- Nathann Cohen (November 2014, written in various airports between Nice and
  Chennai).

Methods
-------
"""
# ****************************************************************************
#       Copyright (C) 2014 Nathann Cohen <nathann.cohen@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  http://www.gnu.org/licenses/
# ****************************************************************************

from libc.stdlib cimport qsort
from libc.stdint cimport uint64_t
from cysignals.memory cimport sig_malloc, sig_calloc, sig_free

ctypedef struct hypergraph:
    int n
    int m
    int limbs
    uint64_t ** sets
    uint64_t * set_space
    int * names

cdef inline int bs_get(uint64_t * bitset, int index) noexcept:
    r"""
    Return a bit of a bitset
    """
    return (bitset[index//64]>>(index%64))&1

cdef inline void bs_set(uint64_t * bitset, int index, int bit) noexcept:
    r"""
    Set a bit of a bitset.

    "bit" *MUST* be equal to either 0 or to 1. The code does not involve any
    "if".
    """
    bitset[index//64] &= ~((<uint64_t> 1)<<index%64)
    bitset[index//64] |= (<uint64_t> bit)<<index%64

cdef inline int bs_issubset64(uint64_t * b1, uint64_t b2, int limbs) noexcept:
    r"""
    Test whether bitset ``b1`` (on ``limbs`` blocks) is a subset of b2 (one block).

    It implies in particular that all last `limbs-1` blocks of ``b1`` are equal
    to zero.
    """
    # Checking that the whole field is zero can be done by computing b1[0] |
    # b1[1] | b[2] | ..., then calling "if" on that variable. May be faster, but
    # I have no real situation on which to test it.
    cdef int i
    for i in range(1,limbs):
        if b1[i]:
            return 0
    return (b1[0]&(~b2)) == 0

cdef void h_free(hypergraph h) noexcept:
    r"""
    Free the hypergraph
    """
    sig_free(h.names)
    sig_free(h.set_space)
    sig_free(h.sets)
    h.names = NULL
    h.set_space = NULL
    h.sets = NULL

cdef hypergraph h_init(int n, list H) noexcept:
    r"""
    Build a C hypergraph from a list `H` of sets on `\{0,...,n-1\}`.
    """
    cdef int x,i
    cdef hypergraph h
    h.n = n
    h.m = len(H)
    h.limbs = (n+63) // 64  # =ceil(n/64)
    h.names = <int *> sig_malloc(sizeof(int)*n)
    h.sets = <uint64_t **> sig_malloc(h.m*sizeof(uint64_t *))
    h.set_space = <uint64_t *> sig_calloc(h.m*(h.limbs+1),sizeof(uint64_t))

    # Consistency check
    for S in H:
        for x in S:
            if x < 0 or x >= n:
                h.n = -1

    if (h.names == NULL or
        h.sets == NULL or
        h.set_space == NULL or
        h.n == -1):
        h.n = -1
        return h

    for i in range(n):
        h.names[i] = i
    for i in range(h.m):
        h.sets[i] = h.set_space+i*(h.limbs+1)

    for i,S in enumerate(H):
        for x in S:
            bs_set(h.sets[i],x,1)
        h.sets[i][h.limbs] = len(S)

    return h

cdef inline void permute(hypergraph * h, int n1, int n2) noexcept:
    r"""
    Permutes two points of h inplace.

    This is only a data structure change, as h still represents the same
    hypergraph. In particular, the names of `n1` and `n2` are also exchanged.
    """
    if n1==n2:
        return

    h.names[n1],h.names[n2] = h.names[n2],h.names[n1]

    cdef int i,b1,b2
    for i in range(h.m):
        b1 = bs_get(h.sets[i],n1)
        b2 = bs_get(h.sets[i],n2)
        bs_set(h.sets[i],n1,b2)
        bs_set(h.sets[i],n2,b1)

cdef induced_hypergraph(hypergraph * h, int n, hypergraph * tmp):
    r"""
    Fills tmp with the hypergraph induced by points {0,...,n-1} in h.

    Assumes `n<=64`.
    """
    cdef int i,num_sets
    num_sets = 0
    cdef uint64_t current_set = ((<uint64_t> 1)<<(n+1))-1
    for i in range(h.m):
        if bs_issubset64(h.sets[i],current_set,h.limbs):
            tmp.sets[num_sets][0] = h.sets[i][0]
            num_sets += 1
    tmp.m = num_sets
    tmp.n = n
    tmp.limbs = 1


cdef void trace_hypergraph64(hypergraph * h, int n, hypergraph * tmp) noexcept:
    r"""
    Store in `tmp` the trace of the sets on `\{0, \ldots, n-1\}` in h1.

    Note that the size of the sets are kept as they are, i.e. the size of a set
    stored in tmp is what it was in h. This is useful information we use to cut
    the exploration.

    Assumes `n<=64`.
    """
    cdef int i
    cdef uint64_t current_set = ((<uint64_t> 1)<<(n+1))-1
    for i in range(h.m):
        tmp.sets[i][0] = h.sets[i][0]&current_set
        tmp.sets[i][1] = h.sets[i][h.limbs]

    tmp.limbs = 1

cdef int is_subhypergraph_admissible(hypergraph h1, hypergraph * h2_trace, int n, hypergraph tmp1) noexcept:
    r"""
    If there are `c` sets of size `k` containing `S\subseteq \{0,...,n-1\}` in
    `h2`, then there must be `>=c` sets of size `k` containing `S` in h1. This
    function checks that this property hold.

    ``h2_trace`` is expected to point toward the trace of ``h2`` on its first
    `n` points. Besides, its sets must be sorted with respect to
    ``cmp_128_bits``.

    Assumes `n<=64`.
    """
    trace_hypergraph64(&h1,n,&tmp1)
    qsort(tmp1.sets,h1.m,sizeof(uint64_t *),cmp_128_bits)

    cdef int i1,i2
    i1 = -1
    for i2 in range(h2_trace.m):
        i1 += 1
        while (i1<h1.m and
               (tmp1.sets[i1][0] < h2_trace.sets[i2][0] or
                tmp1.sets[i1][1] < h2_trace.sets[i2][1])):
            i1 += 1
        if (i1>=h1.m or
            (tmp1.sets[i1][0] > h2_trace.sets[i2][0] or
             tmp1.sets[i1][1] > h2_trace.sets[i2][1])):
            return 0

    return 1

cdef int cmp_128_bits(const void * a, const void * b) noexcept nogil:
    r"""
    Lexicographic order on 128-bits words
    """
    cdef uint64_t * p1 = (<uint64_t **> a)[0]
    cdef uint64_t * p2 = (<uint64_t **> b)[0]
    if p1[0] > p2[0]:
        return 1
    elif p1[0] == p2[0]:
        return 1 if p1[1] > p2[1] else -1
    else:
        return -1

cdef int is_induced_admissible64(hypergraph h1, hypergraph * h2_induced, int n, hypergraph tmp1) noexcept:
    r"""
    Test if the hypergraph induced in h1 by 0,...,n-1 is equal to the hypergraph
    induced in h2 by 0,...,n-1.

    ``h2_induced`` is expected to be a pointer toward the hypergraph induced by
    the first n points of ``h2``. Its sets must be sorted according to
    ``cmp_128_bits``.

    Assumes `n<=64`.
    """
    # *MUST* cache the info of tmp2 which only depends on n !!!!
    # *MUST* cache the sets' size
    induced_hypergraph(&h1,n,&tmp1)

    if tmp1.m!=h2_induced.m:
        return 0

    qsort(tmp1.sets,tmp1.m,sizeof(uint64_t *),cmp_128_bits)

    cdef int i
    for i in range(tmp1.m):
        if tmp1.sets[i][0] != h2_induced.sets[i][0]:
            return 0

    return 1

cdef class SubHypergraphSearch:

    cdef hypergraph h1,h2,tmp1,tmp2
    cdef list points1,points2
    cdef int induced
    cdef int * step
    cdef hypergraph * h2_traces
    cdef hypergraph * h2_induced

    def __cinit__(self, H1, H2, induced):
        r"""
        See the documentation's class.

        EXAMPLES::

            sage: from sage.combinat.designs.subhypergraph_search import SubHypergraphSearch
            sage: g1 = IncidenceStructure(graphs.PetersenGraph().edges(sort=True, labels=False))
            sage: g2 = IncidenceStructure(graphs.CycleGraph(5).edges(sort=True, labels=False))
            sage: S = SubHypergraphSearch(g1,g2,0)
            sage: sum(1 for _ in S)
            120
        """
        self.points1 = H1._points
        self.points2 = H2._points
        self.induced = induced
        cdef int n1 = H1.num_points()
        cdef int n2 = H2.num_points()

        if n2>64:
            raise RuntimeError("H2 has {}>64 points".format(n2))

        self.h1   = h_init(n1,H1._blocks)
        self.h2   = h_init(n2,H2._blocks)
        self.tmp1 = h_init(n1,H1._blocks) # No actual need to fill them,
        self.tmp2 = h_init(n2,H2._blocks) # only allocate the memory

        self.step = <int *> sig_malloc((n2+1)*sizeof(int))

        # all possible traces/induced subgraphs for h2
        #
        # (calloc sets all internal pointers to NULL)
        self.h2_traces  = <hypergraph *> sig_calloc(n2+1,sizeof(hypergraph))
        self.h2_induced = <hypergraph *> sig_calloc(n2+1,sizeof(hypergraph))

        if (self.h1.n   == -1 or
            self.h2.n   == -1 or
            self.tmp1.n == -1 or
            self.tmp2.n == -1 or
            self.h2_traces  == NULL or
            self.h2_induced == NULL or
            self.step       == NULL):
                raise MemoryError # also calls __dealloc__

        self.relabel_heuristic()

        # cache all n2+1 traces of h2, we will need them often.
        cdef int i
        for i in range(n2+1):
            self.h2_traces[i] = h_init(n2,H2._blocks)
            if self.h2_traces[i].n == -1:
                raise MemoryError
            trace_hypergraph64(&self.h2,i,&self.h2_traces[i])
            qsort(self.h2_traces[i].sets,self.h2.m,sizeof(uint64_t *),cmp_128_bits)

        # cache all n2+1 induced subhypergraphs of h2, we will need them often.
        for i in range(n2+1):
            self.h2_induced[i] = h_init(n2,H2._blocks)
            if self.h2_induced[i].n == -1:
                raise MemoryError
            induced_hypergraph(&self.h2,i,&self.h2_induced[i])
            qsort(self.h2_induced[i].sets,self.h2_induced[i].m,sizeof(uint64_t *),cmp_128_bits)

    def __dealloc__(self):
        r"""
        Free the ressources
        """
        # If self.h2 was not allocated, then self.h2.n = -1
        cdef int i
        for i in range(self.h2.n+1):
            h_free(self.h2_traces[i])
            h_free(self.h2_induced[i])
        h_free(self.h1)
        h_free(self.h2)
        h_free(self.tmp1)
        h_free(self.tmp2)
        sig_free(self.step)
        sig_free(self.h2_traces)
        sig_free(self.h2_induced)

    def relabel_heuristic(self):
        r"""
        Relabel `H_2` in order to make the algorithm faster.

        Objective: we try to pick an ordering `p_1,...,p_k` of the points of
        `H_2` that maximizes the number of sets involving the first points in
        the ordering. One way to formalize the problems indicates that it may be
        NP-Hard (generalizes the max clique problem for graphs) so we do not try
        to solve it exactly: we just need a sufficiently good heuristic.

        Assuming that the first points are `p_1,...,p_k`, we determine `p_{k+1}`
        as the point `x` such that the number of sets `S` with `x\in S` and
        `S\cap \{p_1,...,p_k\}\neq \emptyset` is maximal. In case of ties, we
        take a point with maximum degree.

        This function is called when an instance of :class:`SubHypergraphSearch`
        is created.

        EXAMPLES::

            sage: d = designs.projective_plane(3)                                       # needs sage.schemes
            sage: d.isomorphic_substructures_iterator(d).relabel_heuristic()            # needs sage.schemes
        """
        cdef hypergraph h2 = self.h2
        cdef int x,y,i
        cdef list degree = [0]*h2.n
        cdef list degree_with_ordered_points = [0]*h2.n
        cdef uint64_t current_set = 0

        # pre-compute the degree of each point
        for x in range(h2.n):
            y = 0
            for i in range(h2.m):
                y += bs_get(h2.sets[i],x)
            degree[x] = y

        # Compute the actual ordering
        for x in range(h2.n):
            y = 0
            for i in range(h2.m):
                if h2.sets[i][0] & current_set:
                    y += bs_get(h2.sets[i],x)
            degree_with_ordered_points[x] = y
            next_point = max((degree_with_ordered_points[i],degree[i],i)
                             for i in range(x,h2.n))[2]
            permute(&h2,x,next_point)
            current_set += 1<<x

    def __iter__(self):
        r"""
        Iterate over all copies of h2 in h1.

        EXAMPLES:

        How many distinct `C_5` in Petersen's graph ? ::

            sage: P = graphs.PetersenGraph()
            sage: C = graphs.CycleGraph(5)
            sage: IP = IncidenceStructure(P.edges(sort=True, labels=False))
            sage: IC = IncidenceStructure(C.edges(sort=True, labels=False))
            sage: sum(1 for _ in IP.isomorphic_substructures_iterator(IC))
            120
        """
        cdef hypergraph h1 = self.h1
        cdef hypergraph h2 = self.h2
        cdef hypergraph tmp1 = self.tmp1
        cdef int * step = self.step

        if h1.n<h2.n or h1.m<h2.m:
            return

        self.step[0] = -1

        # n is the current point of H2 whose representant must be decided
        cdef int n = 0

        while n>=0:
            step[n] += 1

            # If we went through all permutations beginning with
            # h1.names[0],...,h1.names[n-1].
            if n+step[n]==h1.n:
                step[n] = -1
                n -= 1
                if n>=0:
                    permute(&h1,n,n+step[n])
                continue

            permute(&h1,n,n+step[n])

            # Filter the assignments. If this 'if' is replaced by 'if True:', the
            # code enumerates all n1!/(n1-n2)! injections from [n2] into [n1].
            if ((not self.induced or is_induced_admissible64(h1,&self.h2_induced[n],n,tmp1)) and
                is_subhypergraph_admissible(h1,&self.h2_traces[n],n,tmp1)):
                n += 1
                step[n] = -1
                if n == h2.n:
                    yield {self.points2[h2.names[i]]:self.points1[h1.names[i]] for i in range(h2.n)}
                    n -= 1
                else:
                    continue

            permute(&h1,n,n+step[n])



# sage_setup: distribution = sagemath-graphs
from libcpp cimport bool
from libcpp.list cimport list as cpplist
from libcpp.vector cimport vector

cdef extern from "modular_decomposition.hpp":
    cdef cppclass SDData:
        void set_from_data(size_t lex_label_offset, const int* sigma,
                           const size_t *xslice_len,
                           const vector[int] *lex_label)

    cdef cppclass md_tree_node:
        bool is_leaf() const
        bool is_prime() const
        bool is_parallel() const
        bool is_series() const
        cpplist[md_tree_node *] children
        # If is_leaf() is true, the attribute 'vertex' contains the id of the
        # corresponding vertex. If is_leaf() is false, the attribute 'vertex'
        # contains the id of a vertex corresponding to any leaf below
        # the node (i.e., 'vertex' contains the id of a vertex belonging to the
        # module corresponding to the node).
        int vertex

    void dealloc_md_tree_nodes_recursively(md_tree_node *)

    cdef md_tree_node * corneil_habib_paul_tedder_inner(const SDData &SD)



# sage_setup: distribution = sagemath-graphs
# cython: binding=True
# distutils: language = c++
r"""
Decomposition by clique minimal separators

This module implements methods related to the decomposition of a graph by clique
minimal separators. See [TY1984]_ and [BPS2010]_ for more details on the
algorithms.

Methods
-------
"""
# ****************************************************************************
# Copyright (C) 2019 David Coudert <david.coudert@inria.fr>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from libcpp.vector cimport vector
from libc.stdint cimport uint32_t
from cysignals.signals cimport sig_on, sig_off, sig_check
from memory_allocator cimport MemoryAllocator

from sage.graphs.base.static_sparse_graph cimport short_digraph
from sage.graphs.base.static_sparse_graph cimport init_short_digraph
from sage.graphs.base.static_sparse_graph cimport free_short_digraph
from sage.graphs.base.static_sparse_graph cimport has_edge

from sage.sets.set import Set
from sage.graphs.traversals cimport maximum_cardinality_search_M_short_digraph


def make_tree(atoms, cliques):
    r"""
    Return a tree of atoms and cliques.

    The atoms are the leaves of the tree and the cliques are the internal
    vertices. The number of atoms is the number of cliques plus one.

    As a clique may appear several times in the list ``cliques``, vertices are
    numbered by pairs `(i, S)`, where `0 \leq i < |atoms| + |cliques|` and `S`
    is either an atom or a clique.

    The root of the tree is the only vertex with even or null degree, i.e., 0 if
    ``cliques`` is empty and 2 otherwise. When ``cliques`` is not empty, other
    internal vertices (each of which is a clique) have degree 3, and the
    leaves (vertices of degree 1) are the atoms.

    INPUT:

    - ``atoms`` -- list of atoms

    - ``cliques`` -- list of cliques

    EXAMPLES::

        sage: from sage.graphs.graph_decompositions.clique_separators import make_tree
        sage: G = graphs.Grid2dGraph(2, 4)
        sage: A, Sc = G.atoms_and_clique_separators()
        sage: T = make_tree(A, Sc)
        sage: all(u[1] in A for u in T if T.degree(u) == 1)
        True
        sage: all(u[1] in Sc for u in T if T.degree(u) != 1)
        True

    TESTS::

        sage: from sage.graphs.graph_decompositions.clique_separators import make_tree
        sage: make_tree([0], [1, 2])
        Traceback (most recent call last):
        ...
        ValueError: the number of atoms must be the number of cliques plus one
    """
    if (atoms or cliques) and len(atoms) != len(cliques) + 1:
        raise ValueError("the number of atoms must be the number of cliques plus one")

    from sage.graphs.graph import Graph
    T = Graph()
    if cliques:
        # As a clique can appear several times, we number the vertices by
        # pairs (int, Set), with 0 <= int < |atoms| + |cliques|
        T.add_path(list(enumerate(cliques)))
        j = len(cliques)
        T.add_edges((s, (i + j, a)) for s, (i, a) in zip(enumerate(cliques), enumerate(atoms)))
        # We have |atoms| = |cliques| + 1. So |atoms| + |cliques| = 2 * j + 1
        T.add_edge((j - 1, cliques[-1]), (2 * j, atoms[-1]))

    elif atoms:
        # The graph has no clique separator
        T.add_vertex(atoms[0])

    return T


def make_labelled_rooted_tree(atoms, cliques):
    r"""
    Return a :class:`~LabelledRootedTree` of atoms and cliques.

    The atoms are the leaves of the tree and the cliques are the internal
    vertices. The number of atoms is the number of cliques plus one.

    EXAMPLES::

        sage: G = graphs.PathGraph(5)
        sage: ascii_art(G.atoms_and_clique_separators(rooted_tree=True))
          _____{3}_____
         /            /
        {3, 4}   ____{2}____
                /          /
               {2, 3}   __{1}__
                       /      /
                      {0, 1} {1, 2}

    TESTS::

        sage: from sage.graphs.graph_decompositions.clique_separators import make_labelled_rooted_tree
        sage: make_labelled_rooted_tree([0], [1, 2])
        Traceback (most recent call last):
        ...
        ValueError: the number of atoms must be the number of cliques plus one
    """
    from sage.combinat.rooted_tree import LabelledRootedTree
    if not atoms and not cliques:
        return LabelledRootedTree([])

    if len(atoms) != len(cliques) + 1:
        raise ValueError("the number of atoms must be the number of cliques plus one")

    def to_tree(i, n):
        if i < n:
            return LabelledRootedTree([LabelledRootedTree([], label=atoms[i]), to_tree(i + 1, n)],
                                      label=cliques[i])
        return LabelledRootedTree([], label=atoms[i])

    return to_tree(0, len(cliques))


cdef inline bint is_clique(short_digraph sd, vector[int] Hx) noexcept:
    """
    Check if the subgraph sd[Hx] is a clique.

    This is a helper function of ``atoms_and_clique_separators``.
    """
    cdef size_t Hx_size = Hx.size()
    cdef size_t i, j
    cdef int u
    for i in range(Hx_size - 1):
        u = Hx[i]
        for j in range(i + 1, Hx_size):
            if not has_edge(sd, u, Hx[j]):
                return False
    return True


def atoms_and_clique_separators(G, tree=False, rooted_tree=False, separators=False):
    r"""
    Return the atoms of the decomposition of `G` by clique minimal separators.

    Let `G = (V, E)` be a graph. A set `S \subset V` is a clique separator if
    `G[S]` is a clique and the graph `G \setminus S` has at least 2 connected
    components. Let `C \subset V` be the vertices of a connected component of `G
    \setminus S`. The graph `G[C + S]` is an *atom* if it has no clique
    separator.

    This method implements the algorithm proposed in [BPS2010]_, that improves
    upon the algorithm proposed in [TY1984]_, for computing the atoms and the
    clique minimal separators of a graph. This algorithm is based on the
    :meth:`~sage.graphs.traversals.maximum_cardinality_search_M` graph traversal
    and has time complexity in `O(|V|\cdot|E|)`.

    If the graph is not connected, we insert empty separators between the lists
    of separators of each connected components. See the examples below for more
    details.

    INPUT:

    - ``G`` -- a Sage graph

    - ``tree`` -- boolean (default: ``False``); whether to return the result as
      a directed tree in which internal nodes are clique separators and leaves
      are the atoms of the decomposition. Since a clique separator is repeated
      when its removal partition the graph into 3 or more connected components,
      vertices are labels by tuples `(i, S)`, where `S` is the set of vertices
      of the atom or the clique separator, and `0 \leq i \leq |T|`.

    - ``rooted_tree`` -- boolean (default: ``False``); whether to return the
      result as a :class:`~sage.combinat.rooted_tree.LabelledRootedTree`. When
      ``tree`` is ``True``, this parameter is ignored.

    - ``separators`` -- boolean (default: ``False``); whether to also return the
      complete list of separators considered during the execution of the
      algorithm. When ``tree`` or ``rooted_tree`` is ``True``, this parameter is
      ignored.

    OUTPUT:

    - By default, return a tuple `(A, S_c)`, where `A` is the list of atoms of
      the graph in the order of discovery, and `S_c` is the list of clique
      separators, with possible repetitions, in the order the separator has been
      considered. If furthermore ``separators`` is ``True``, return a tuple `(A,
      S_h, S_c)`, where `S_c` is the list of considered separators of the graph
      in the order they have been considered.

    - When ``tree`` is ``True``, format the result as a directed tree

    - When ``rooted_tree`` is ``True`` and ``tree`` is ``False``, format the
      output as a :class:`~sage.combinat.rooted_tree.LabelledRootedTree`

    EXAMPLES:

    Example of [BPS2010]_::

        sage: G = Graph({'a': ['b', 'k'], 'b': ['c'], 'c': ['d', 'j', 'k'],
        ....:            'd': ['e', 'f', 'j', 'k'], 'e': ['g'],
        ....:            'f': ['g', 'j', 'k'], 'g': ['j', 'k'], 'h': ['i', 'j'],
        ....:            'i': ['k'], 'j': ['k']})
        sage: atoms, cliques = G.atoms_and_clique_separators()
        sage: sorted(sorted(a) for a in atoms)
        [['a', 'b', 'c', 'k'],
         ['c', 'd', 'j', 'k'],
         ['d', 'e', 'f', 'g', 'j', 'k'],
         ['h', 'i', 'j', 'k']]
        sage: sorted(sorted(c) for c in cliques)
        [['c', 'k'], ['d', 'j', 'k'], ['j', 'k']]
        sage: T = G.atoms_and_clique_separators(tree=True)
        sage: T.is_tree()
        True
        sage: T.diameter() == len(atoms)
        True
        sage: all(u[1] in atoms for u in T if T.degree(u) == 1)
        True
        sage: all(u[1] in cliques for u in T if T.degree(u) != 1)
        True

    A graph without clique separator::

        sage: G = graphs.CompleteGraph(5)
        sage: G.atoms_and_clique_separators()
        ([{0, 1, 2, 3, 4}], [])
        sage: ascii_art(G.atoms_and_clique_separators(rooted_tree=True))
        {0, 1, 2, 3, 4}

    Graphs with several biconnected components::

        sage: G = graphs.PathGraph(4)
        sage: ascii_art(G.atoms_and_clique_separators(rooted_tree=True))
          ____{2}____
         /          /
        {2, 3}   __{1}__
                /      /
               {1, 2} {0, 1}

        sage: G = graphs.WindmillGraph(3, 4)
        sage: G.atoms_and_clique_separators()
        ([{0, 1, 2}, {0, 3, 4}, {0, 5, 6}, {0, 8, 7}], [{0}, {0}, {0}])
        sage: ascii_art(G.atoms_and_clique_separators(rooted_tree=True))
          ________{0}________
         /                  /
        {0, 1, 2}   _______{0}______
                   /               /
                  {0, 3, 4}   ____{0}___
                             /         /
                            {0, 8, 7} {0, 5, 6}

    When the removal of a clique separator results in `k > 2` connected
    components, this separator is repeated `k - 1` times, but the repetitions
    are not necessarily contiguous::

        sage: G = Graph(2)
        sage: for i in range(5):
        ....:     G.add_cycle([0, 1, G.add_vertex()])
        sage: ascii_art(G.atoms_and_clique_separators(rooted_tree=True))
          _________{0, 1}_____
         /                   /
        {0, 1, 4}   ________{0, 1}_____
                   /                  /
                  {0, 1, 2}   _______{0, 1}___
                             /               /
                            {0, 1, 3}   ____{0, 1}
                                       /         /
                                      {0, 1, 5} {0, 1, 6}

        sage: G = graphs.StarGraph(3)
        sage: G.subdivide_edges(G.edges(sort=False), 2)
        sage: ascii_art(G.atoms_and_clique_separators(rooted_tree=True))
          ______{5}______
         /              /
        {1, 5}   ______{7}______
                /              /
               {2, 7}   ______{9}______
                       /              /
                      {9, 3}   ______{6}______
                              /              /
                             {6, 7}   ______{4}_____
                                     /             /
                                    {4, 5}   _____{0}_____
                                            /            /
                                           {0, 6}   ____{8}____
                                                   /          /
                                                  {8, 9}   __{0}__
                                                          /      /
                                                         {0, 8} {0, 4}

    If the graph is not connected, we insert empty separators between the lists
    of separators of each connected components. For instance, let `G` be a graph
    with 3 connected components. The method returns the list `S_c =
    [S_0,\cdots,S_{i},\ldots, S_{j},\ldots,S_{k-1}]` of `k` clique separators,
    where `i` and `j` are the indexes of the inserted empty separators and `0
    \leq i < j < k - 1`. The method also returns the list `A =
    [A_0,\ldots,S_{k}]` of the `k + 1` atoms, with `k + 1 \geq 3`. The lists of
    atoms and clique separators of each of the connected components are
    respectively `[A_0,\ldots,A_{i}]` and `[S_0,\ldots,S_{i-1}]`,
    `[A_{i+1},\ldots,A_{j}]` and `[S_{i+1},\ldots,S_{j-1}]`, and
    `[A_{j+1},\ldots,A_{k}]` and `[S_{j+1},\ldots,S_{k-1}]`. One can check that
    for each connected component, we get one atom more than clique separators::

        sage: G = graphs.PathGraph(3) * 3
        sage: A, Sc = G.atoms_and_clique_separators()
        sage: A
        [{1, 2}, {0, 1}, {4, 5}, {3, 4}, {8, 7}, {6, 7}]
        sage: Sc
        [{1}, {}, {4}, {}, {7}]
        sage: i , j = [i for i, s in enumerate(Sc) if not s]
        sage: i, j
        (1, 3)
        sage: A[:i+1], Sc[:i]
        ([{1, 2}, {0, 1}], [{1}])
        sage: A[i+1:j+1], Sc[i+1:j]
        ([{4, 5}, {3, 4}], [{4}])
        sage: A[j+1:], Sc[j+1:]
        ([{8, 7}, {6, 7}], [{7}])
        sage: I = [-1, i, j, len(Sc)]
        sage: for i, j in zip(I[:-1], I[1:]):
        ....:     print(A[i+1:j+1], Sc[i+1:j])
        [{1, 2}, {0, 1}] [{1}]
        [{4, 5}, {3, 4}] [{4}]
        [{8, 7}, {6, 7}] [{7}]
        sage: ascii_art(G.atoms_and_clique_separators(rooted_tree=True))
          ______{1}______
         /              /
        {1, 2}   ______{}______
                /             /
               {0, 1}   _____{4}_____
                       /            /
                      {4, 5}   ____{}_____
                              /          /
                             {3, 4}   __{7}__
                                     /      /
                                    {6, 7} {8, 7}

    Loops and multiple edges are ignored::

        sage: G.allow_loops(True)
        sage: G.add_edges([(u, u) for u in G])
        sage: G.allow_multiple_edges(True)
        sage: G.add_edges(G.edges(sort=False))
        sage: ascii_art(G.atoms_and_clique_separators(rooted_tree=True))
          ______{1}______
         /              /
        {1, 2}   ______{}______
                /             /
               {0, 1}   _____{4}_____
                       /            /
                      {4, 5}   ____{}_____
                              /          /
                             {3, 4}   __{7}__
                                     /      /
                                    {6, 7} {8, 7}

    We can check that the returned list of separators is valid::

        sage: G = graphs.RandomGNP(50, .1)
        sage: while not G.is_connected():
        ....:     G = graphs.RandomGNP(50, .1)
        sage: _, separators, _ = G.atoms_and_clique_separators(separators=True)
        sage: for S in separators:
        ....:     H = G.copy()
        ....:     H.delete_vertices(S)
        ....:     if H.is_connected():
        ....:         raise ValueError("something goes wrong")

    TESTS::

        sage: EmptyGraph = Graph()
        sage: EmptyGraph.atoms_and_clique_separators()
        ([], [])
        sage: EmptyGraph.atoms_and_clique_separators(separators=True)
        ([], [], [])
        sage: EmptyGraph.atoms_and_clique_separators(tree=True)
        Graph on 0 vertices
        sage: EmptyGraph.atoms_and_clique_separators(rooted_tree=True)
        None[]
        sage: ascii_art(EmptyGraph.atoms_and_clique_separators(rooted_tree=True))
        None
        sage: I4 = Graph(4)
        sage: I4.atoms_and_clique_separators()
        ([{0}, {1}, {2}, {3}], [{}, {}, {}])
        sage: I4.atoms_and_clique_separators(separators=True)
        ([{0}, {1}, {2}, {3}], [{}, {}, {}], [{}, {}, {}])
        sage: I4.atoms_and_clique_separators(tree=True)
        Graph on 7 vertices
        sage: I4.atoms_and_clique_separators(rooted_tree=True)
        {}[{0}[], {}[{1}[], {}[{3}[], {2}[]]]]
        sage: ascii_art(I4.atoms_and_clique_separators(rooted_tree=True))
          ___{}___
         /       /
        {0}   __{}___
             /      /
            {1}   _{}_
                 /   /
                {3} {2}
    """
    cdef list A = []   # atoms
    cdef list Sh = []  # separators
    cdef list Sc = []  # clique separators
    cdef bint first = True

    if not G.is_connected():
        from sage.graphs.graph import Graph

        for cc in G.connected_components(sort=False):
            g = Graph([cc, G.edge_boundary(cc, cc, False, False)],
                      format='vertices_and_edges',
                      loops=True, multiedges=True)
            res = g.atoms_and_clique_separators(tree=False, rooted_tree=False, separators=separators)

            # Update lists of atoms, separators and clique separators
            A.extend(res[0])
            if separators:
                if not first:
                    Sh.append(Set())
                Sh.extend(res[1])
            if not first:
                Sc.append(Set())
            if separators:
                Sc.extend(res[2])
            else:
                Sc.extend(res[1])
            first = False

        # Format and return the result
        if tree:
            return make_tree(A, Sc)
        elif rooted_tree:
            return make_labelled_rooted_tree(A, Sc)
        elif separators:
            return A, Sh, Sc
        return A, Sc

    cdef int N = G.order()
    cdef list int_to_vertex = list(G)

    # Copying the whole graph to obtain the list of neighbors quicker than by
    # calling out_neighbors. This data structure is well documented in the
    # module sage.graphs.base.static_sparse_graph
    cdef short_digraph sd
    init_short_digraph(sd, G, edge_labelled=False, vertex_list=int_to_vertex)

    # variables for the manipulation of the short digraph
    cdef uint32_t** p_vertices = sd.neighbors
    cdef uint32_t* p_tmp
    cdef uint32_t* p_end

    cdef MemoryAllocator mem = MemoryAllocator()
    cdef int* alpha = <int*>mem.calloc(N, sizeof(int))
    cdef int* alpha_inv = <int*>mem.calloc(N, sizeof(int))
    cdef bint* X = <bint*>mem.calloc(N, sizeof(bint))
    cdef bint* active = <bint*>mem.calloc(N, sizeof(bint))
    cdef int* waiting_list = <int*>mem.calloc(N, sizeof(int))
    cdef int* seen = <int*>mem.calloc(N, sizeof(int))
    cdef list F = []
    cdef int u, v, waiting_begin, waiting_end

    sig_on()
    maximum_cardinality_search_M_short_digraph(sd, 0, alpha, alpha_inv, F, X)
    sig_off()

    # Instead of building the graph H of the triangulation, extracting the
    # neighbors of vertex alpha[i] and then removing that vertex from H, we
    # directly build the neighborhoods. Note that vertices are removed from the
    # graph of the triangulation in the order of alpha. Hence, at step i, this
    # graph has no vertex u such that alpha_inv[u] < i, and edge (u, v) is
    # removed from the graph when min(alpha_inv[u], alpha_inv[v]) is removed.
    # The neighborhood of x at step i is thus {v in N_H(x) | alpha_inv[v] > i}.
    cdef vector[vector[int]] H
    H.resize(N)
    for u in range(N):
        p_tmp = p_vertices[u]
        p_end = p_vertices[u + 1]
        while p_tmp < p_end:
            v = p_tmp[0]
            p_tmp += 1
            if u < v:
                # We consider edge (u, v) only once, when u > v, and the
                # short_digraph data structure ensures that neighbors are sorted
                break
            if alpha_inv[u] < alpha_inv[v]:
                if X[u]:
                    H[u].push_back(v)
            elif X[v]:
                H[v].push_back(u)
    for u, v in F:
        if alpha_inv[u] < alpha_inv[v]:
            if X[u]:
                H[u].push_back(v)
        elif X[v]:
            H[v].push_back(u)

    # Instead of using a copy Gp of G and removing from it the vertices of the
    # connected component of an atom after its discovery, we use an array of
    # booleans to avoid visiting inactive vertices
    for u in range(N):
        active[u] = True
        seen[u] = -1

    cdef frozenset Sint
    cdef vector[int] Sint_min
    cdef vector[int] Cint
    cdef vector[int] Hx

    for i in range(N):
        sig_check()
        x = alpha[i]
        if X[x] and not H[x].empty():
            Hx = H[x]

            if separators:
                Sh.append(Set(int_to_vertex[u] for u in Hx))

            if is_clique(sd, Hx):
                # The subgraph G[H[x]] = G[S] is a clique.
                # We extract the connected component of Gp - S containing x
                Sint = frozenset(Hx)
                Sint_min.clear()
                Cint.clear()
                Cint.push_back(x)
                seen[x] = x
                waiting_list[0] = x
                waiting_begin = 0
                waiting_end = 0
                while waiting_begin <= waiting_end:

                    u = waiting_list[waiting_begin]
                    waiting_begin += 1
                    p_tmp = p_vertices[u]
                    end = p_vertices[u + 1]

                    while p_tmp < end:
                        v = p_tmp[0]
                        p_tmp += 1

                        if active[v] and seen[v] != x:
                            seen[v] = x
                            if v in Sint:
                                # We keep only the vertices of the clique
                                # separator incident to the connected component
                                # containing x
                                Sint_min.push_back(v)
                            else:
                                Cint.push_back(v)
                                waiting_end += 1
                                waiting_list[waiting_end] = v

                # Store the atom Smin + C and the minimal clique separator Smin
                Smin = Set(int_to_vertex[u] for u in Sint_min)
                A.append(Set(Smin.set().union(int_to_vertex[u] for u in Cint)))
                Sc.append(Smin)

                # "Remove" the vertices of Cint from the graph Gp
                for u in Cint:
                    active[u] = False

    free_short_digraph(sd)
    H.clear()

    # We add the last atom
    if Sc:
        A.append(Set(int_to_vertex[x] for x in range(N) if active[x]))
    elif G:
        # The graph has no clique separator
        A.append(Set(int_to_vertex))

    # Format and return the result
    if tree:
        return make_tree(A, Sc)
    elif rooted_tree:
        return make_labelled_rooted_tree(A, Sc)
    if separators:
        return A, Sh, Sc
    return A, Sc



# sage_setup: distribution = sagemath-graphs
# sage.doctest: needs sage.graphs
r"""
Finite filtered complexes

AUTHORS:

- Guillaume Rousseau (2021-05)

This module implements the basic structures of finite filtered complexes.
A filtered complex is a simplicial complex, where each simplex is given
a weight, or "filtration value", such that the weight of a simplex is
greater than the weight of each of its faces.

The algorithm used in this module comes from [ZC2005]_.

EXAMPLES::

    sage: FilteredSimplicialComplex([([0], 0), ([1], 0), ([0, 1], 1)])
    Filtered complex on vertex set (0, 1) and
     with simplices ((0,) : 0), ((1,) : 0), ((0, 1) : 1)

Sage can compute persistent homology of simplicial complexes::

    sage: X = FilteredSimplicialComplex([([0], 0), ([1], 0), ([0, 1], 1)])
    sage: X.persistence_intervals(0)                                                    # needs sage.modules
    [(0, 1), (0, +Infinity)]

FilteredSimplicialComplex objects are mutable. Filtration values can be
set with the ``filtration`` method as follows::

    sage: X = FilteredSimplicialComplex() # returns an empty complex
    sage: X.persistence_intervals(1)                                                    # needs sage.modules
    []
    sage: X.filtration(Simplex([0, 2]), 0) # recursively adds faces
    sage: X.filtration(Simplex([0, 1]), 0)
    sage: X.filtration(Simplex([1, 2]), 0)
    sage: X.filtration(Simplex([0, 1, 2]), 1) # closes the circle
    sage: X.persistence_intervals(1)                                                    # needs sage.modules
    [(0, 1)]

The filtration value of a simplex can be accessed as well with the
``filtration`` method, by not specifying a filtration value in
the arguments. If the simplex is not in the complex, this returns
``None``::

    sage: X = FilteredSimplicialComplex([([0], 0), ([1], 0), ([0,1], 1)])
    sage: X.filtration(Simplex([0]))
    0
    sage: X.filtration(Simplex([1,2])) is None
    True

Filtration values can be accessed with function call and list
syntax as follows::

    sage: X = FilteredSimplicialComplex([([0], 0), ([1], 0), ([0,1], 1)])
    sage: s_1 = Simplex([0])
    sage: X[s_1]
    0
    sage: X(Simplex([0,1]))
    1
    sage: X(Simplex(['baba']))
    <BLANKLINE>

It is also possible to set the filtration value of a simplex with
the ``insert`` method, which takes as argument a list of vertices
rather than a ``Simplex``. This can make code more readable / clear::

    sage: X = FilteredSimplicialComplex()
    sage: X.insert(['a'], 0)
    sage: X.insert(['b', 'c'], 1)
    sage: X
    Filtered complex on vertex set ('a', 'b', 'c') and with simplices
     (('a',) : 0), (('c',) : 1), (('b',) : 1), (('b', 'c') : 1)
"""

# ****************************************************************************
#       Copyright (C) 2013 GUILLAUME ROUSSEAU <guillaume.rousseau@ens-lyon.fr>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.misc.cachefunc import cached_method
from sage.misc.lazy_import import lazy_import
from sage.rings.integer import Integer
from sage.rings.infinity import infinity
from sage.structure.sage_object import SageObject
from sage.topology.simplicial_complex import Simplex, SimplicialComplex

lazy_import('sage.modules.free_module', 'FreeModule')
lazy_import('sage.rings.finite_rings.finite_field_constructor', 'GF')


class FilteredSimplicialComplex(SageObject):
    r"""
    Define a filtered complex.

    INPUT:

    - ``simplices`` -- list of simplices and filtration values
    - ``verbose`` -- boolean (default: ``False``); if ``True``, any change to
      the filtration value of a simplex will be printed

    ``simplices`` should be a list of tuples ``(l, v)``, where
    ``l`` is a list of vertices and ``v`` is the corresponding
    filtration value.

    EXAMPLES::

        sage: FilteredSimplicialComplex([([0], 0), ([1], 0), ([2], 1), ([0,1], 2.27)])
        Filtered complex on vertex set (0, 1, 2) and with simplices
         ((0,) : 0), ((1,) : 0), ((2,) : 1), ((0, 1) : 2.27000000000000)
    """
    def __init__(self, simplices=[], verbose=False):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: X = FilteredSimplicialComplex([([0], 0), ([1], 0), ([2], 1), ([0,1], 2.27)])
            sage: TestSuite(X).run()
        """
        # _vertices is the set of vertices on which the complex
        # is constructed
        self._vertices = set()

        # _filtration_dict has simplices as keys
        # and entries are corresponding filtration values
        self._filtration_dict = {}
        self._dimension = 0
        self._max_value = 0

        # when _verbose is set to True, insertion
        # will warn the user when something non-trivial
        # happens.
        self._verbose = verbose

        # Insert all simplices in the initial list
        for l, v in simplices:
            self.insert(l, v)

    def __eq__(self, other):
        """
        Check equality.

        EXAMPLES::

            sage: X = FilteredSimplicialComplex([([0], 0), ([1], 0), ([2], 1), ([0,1], 2.27)])
            sage: Y = FilteredSimplicialComplex()
            sage: Y.insert([0], 0)
            sage: Y.insert([1], 0)
            sage: Y.insert([2], 1)
            sage: Y.insert([0,1], 2.27)
            sage: X == Y
            True
            sage: Y.filtration([1,2], 2)
            sage: X == Y
            False

            sage: Y = FilteredSimplicialComplex([([0], 0), ([1], 0), ([2], 1), ([0,1], 2)])
            sage: X == Y
            False
        """
        return (isinstance(other, FilteredSimplicialComplex)
                and self._vertices == other._vertices
                and self._filtration_dict == other._filtration_dict)

    def __ne__(self, other):
        """
        Check inequality.

        EXAMPLES::

            sage: X = FilteredSimplicialComplex([([0], 0), ([1], 0), ([2], 1), ([0,1], 2.27)])
            sage: Y = FilteredSimplicialComplex([([0], 0), ([1], 0), ([2], 1), ([0,1], 3)])
            sage: X != Y
            True
            sage: Y.filtration([0,1], 2.27)
            sage: X != Y
            False
        """
        return not (self == other)

    def _get_value(self, s):
        r"""
        Return the filtration value of a simplex ``s`` in the complex.

        EXAMPLES::

            sage: X = FilteredSimplicialComplex([([0], 1), ([1], 2)])
            sage: X._get_value(Simplex([0]))
            1

        This also works for the call and getitem syntax as a shorthand::

            sage: X(Simplex([0]))
            1
            sage: X[Simplex([0])]
            1
        """
        if s in self._filtration_dict:
            return self._filtration_dict[s]
        else:
            return None

    __call__ = _get_value
    __getitem__ = _get_value

    def _insert(self, simplex, filtration_value):
        r"""
        Add a simplex to the complex.

        All faces of the simplex are added recursively if they are
        not already present, with the same value.
        If the simplex is already present, and the new value is lower
        than its current value in the complex, the value gets updated,
        otherwise it does not change. This propagates recursively to faces.

        If verbose has been enabled, this method will describe what it
        is doing during an insertion.

        INPUT:

        - ``simplex`` -- :class:`Simplex`; simplex to be inserted
        - ``filtration_value`` -- value of the simplex

        EXAMPLES::

            sage: X = FilteredSimplicialComplex()
            sage: X._insert(Simplex([0]), 3)
            sage: X
            Filtered complex on vertex set (0,) and with simplices ((0,) : 3)
        """
        # Keep track of whether the simplex is already in the complex
        # and if it should be updated or not
        curr_value = self[simplex]
        if curr_value is not None:
            if self._verbose:
                print("Face {} is already in the complex.".format(simplex))
            if curr_value > filtration_value:
                if self._verbose:
                    verbose_string = "However its value is {}".format(curr_value)
                    verbose_string += ": updating it to {}".format(filtration_value)
                    print(verbose_string)
                self._filtration_dict.pop(simplex)
            else:
                if self._verbose:
                    print("Its value is {}: keeping it that way".format(curr_value))
                return

        # check that all faces are in the complex already.
        # If not, warn the user and add faces (recursively)
        faces = simplex.faces()
        if simplex.dimension() > 0:
            for f in faces:
                if self._verbose:
                    print("Also inserting face {} with value {}".format(f, filtration_value))
                self._insert(f, filtration_value)

        self._filtration_dict[simplex] = filtration_value
        self._dimension = max(self._dimension, simplex.dimension())
        self._max_value = max(self._max_value, filtration_value)
        self._vertices.update(simplex.set())
        self._persistent_homology.clear_cache()

    def insert(self, vertex_list, filtration_value):
        r"""
        Add a simplex to the complex.

        All faces of the simplex are added recursively if they are
        not already present, with the same value.
        If the simplex is already present, and the new value is lower
        than its current value in the complex, the value gets updated,
        otherwise it does not change. This propagates recursively to faces.

        If verbose has been enabled, this method will describe what it
        is doing during an insertion.

        INPUT:

        - ``vertex_list`` -- list of vertices
        - ``filtration_value`` -- desired value of the simplex to be added

        EXAMPLES::

            sage: X = FilteredSimplicialComplex()
            sage: X.insert(Simplex([0]),3)
            sage: X
            Filtered complex on vertex set (0,) and with simplices ((0,) : 3)

        If the verbose parameter was set to true, this method will print
        some info::

            sage: X = FilteredSimplicialComplex(verbose=True)
            sage: X.insert(Simplex([0, 1]), 2)
            Also inserting face (1,) with value 2
            Also inserting face (0,) with value 2
            sage: X.insert(Simplex([0]),1)
            Face (0,) is already in the complex.
            However its value is 2: updating it to 1
            sage: X.insert(Simplex([0]), 77)
            Face (0,) is already in the complex.
            Its value is 1: keeping it that way
        """
        self._insert(Simplex(vertex_list), filtration_value)

    def filtration(self, s, filtration_value=None):
        r"""
        Set filtration value of a simplex, or return value
        of existing simplex.

        INPUT:

        - ``s`` -- :class:`Simplex` for which to set or obtain the
          value of
        - ``filtration_value`` -- (optional) filtration value
          for the simplex

        If no filtration value is specified, this returns the value of
        the simplex in the complex. If the simplex is not in the complex,
        this returns ``None``.

        If ``filtration_value`` is set, this function inserts the
        simplex into the complex with the specified value.
        See documentation of :meth:`insert` for more details.

        EXAMPLES::

            sage: X = FilteredSimplicialComplex([([0], 0), ([1], 1)])
            sage: X.filtration(Simplex([0, 1])) is None
            True
            sage: X.filtration(Simplex([0, 1]), 2)
            sage: X.filtration([0, 1])
            2
        """
        s = Simplex(s)
        if filtration_value is None:
            return self._get_value(s)
        else:
            self._insert(s, filtration_value)

    def prune(self, threshold):
        r"""
        Return a copy of the filtered complex, where simplices above
        the threshold value have been removed.

        INPUT:

        - ``threshold`` -- a real value, above which simplices are discarded

        Simplices with filtration value exactly equal to ``threshold``
        are kept in the result.

        EXAMPLES::

            sage: a = FilteredSimplicialComplex()
            sage: a.insert([0], 0)
            sage: a.insert([0, 1], 1)
            sage: a.insert([0, 2], 2)
            sage: b = a.prune(1)
            sage: b
            Filtered complex on vertex set (0, 1) and
             with simplices ((0,) : 0), ((1,) : 1), ((0, 1) : 1)
        """
        result_complex = FilteredSimplicialComplex()
        for s in self._filtration_dict:
            if self[s] <= threshold:
                result_complex._insert(s, self[s])

        return result_complex

    @cached_method(key=lambda self, f, s, v: (f, s))
    def _persistent_homology(self, field=2, strict=True, verbose=False):
        """
        Compute the homology intervals of the complex.

        INPUT:

        - ``field`` -- (default: 2) prime number modulo which the homology
          is computed
        - ``strict`` -- boolean (default: ``True``); if ``False``, takes into account
            intervals of persistence 0
        - ``verbose`` -- boolean (default: ``False``); if ``True``, prints the
          progress of computation

        This method is called whenever Betti numbers or intervals are
        computed, and the result is cached. It returns the list of
        intervals.

        ALGORITHM:

        The computation behind persistent homology is a matrix reduction.
        The algorithm implemented is described in [ZC2005]_.

        EXAMPLES::

            sage: X = FilteredSimplicialComplex([([0], 0), ([1], 0), ([0,1], 2)])
            sage: X._persistent_homology()[0]                                           # needs sage.modules
            [(0, 2), (0, +Infinity)]

        Some homology elements may have a lifespan or persistence of 0.
        They are usually discarded, but can be kept if necessary::

            sage: X = FilteredSimplicialComplex()
            sage: X.insert([0,1],1) # opens a hole and closes it instantly
            sage: X._persistent_homology(strict=False)[0]                               # needs sage.modules
            [(1, 1), (1, +Infinity)]

        REFERENCES:

        - [ZC2005]_

        TESTS:

        This complex is used as a running example in [ZC2005]_::

            sage: l = [([0], 0), ([1], 0), ([2], 1), ([3], 1), ([0, 1], 1),
            ....:      ([1, 2], 1), ([0, 3], 2), ([2, 3], 2), ([0, 2], 3),
            ....:      ([0, 1, 2], 4), ([0, 2, 3], 5)]
            sage: X = FilteredSimplicialComplex(l)
            sage: X.persistence_intervals(0)                                            # needs sage.modules
            [(0, 1), (1, 2), (0, +Infinity)]
            sage: X.persistence_intervals(1)                                            # needs sage.modules
            [(3, 4), (2, 5)]
            sage: X.persistence_intervals(0, strict=False)                              # needs sage.modules
            [(0, 1), (1, 1), (1, 2), (0, +Infinity)]
        """
        # first, order the simplices in lexico order
        # on dimension, value and then arbitrary order
        # defined by the Simplex class.
        def key(s):
            d = self._get_value(s)
            return (s.dimension(), d, s)
        simplices = list(self._filtration_dict)
        simplices.sort(key=key)

        # remember the index of each simplex in a dict
        self._index_of_simplex = {}
        n = len(simplices)
        for i in range(n):
            self._index_of_simplex[simplices[i]] = i

        self._field_int = field
        self._field = GF(field)
        self._chaingroup = FreeModule(self._field, rank_or_basis_keys=simplices)

        # Initialize data structures for the algo
        self._marked = [False] * n
        self._T = [None] * n
        intervals = [[] for i in range(self._dimension + 1)]
        self.pairs = []

        self._strict = strict
        self._verbose = verbose

        if self._verbose:
            print("Beginning first pass")

        for j in range(n):
            # if being verbose, print progress
            # every 1000 simplices.
            if self._verbose and j % 1000 == 0:
                print('{}/{}'.format(j, n))

            s = simplices[j]
            d = self._remove_pivot_rows(s, simplices)

            if d == 0:
                self._marked[j] = True
            else:
                max_index = self._max_index(d)
                t = simplices[max_index]
                self._T[max_index] = (s, d)
                self._add_interval(t, s, intervals)

        if self._verbose:
            print("First pass over, beginning second pass")

        for j in range(n):
            if self._verbose and j % 1000 == 0:
                print('{}/{}'.format(j, n))

            s = simplices[j]
            if self._marked[j] and not self._T[j]:
                self._add_interval(s, None, intervals)

        if self._verbose:
            print("Second pass over")
        return intervals

    def _add_interval(self, s, t, intervals):
        r"""
        Add a new interval (i.e. homology element).

        This method should not be called by users, it is used in
        the :meth:`_persistent_homology` method. The simplex of
        death may be ``None``, in which case the interval is infinite.

        INPUT:

        - ``s`` -- birth simplex
        - ``t`` -- death simplex
        - ``intervals`` -- list of current intervals

        If ``t`` is not ``None``, its dimension should be
        one more than the dimension of ``s``.

        TESTS::

            sage: # needs sage.modules
            sage: X = FilteredSimplicialComplex([([0], 0), ([1, 2], 10)])
            sage: int_list = X._persistent_homology()
            sage: int_list[0]
            [(0, +Infinity), (10, +Infinity)]
            sage: X._add_interval(Simplex([0]), Simplex([1, 2]), int_list)
            sage: int_list[0]
            [(0, +Infinity), (10, +Infinity), (0, 10)]

        Infinite interval::

            sage: int_list2 = [[],[]]
            sage: X._add_interval(Simplex([1, 2]), None, int_list2)                     # needs sage.modules
            sage: int_list2[1]
            [(10, +Infinity)]
        """
        # figure out dimension of homology element
        # and indices of the two simplices. If the
        # closing simplex is None, then the interval
        # is infinite.
        k = s.dimension()
        i = self._filtration_dict[s]
        if not t:
            j = infinity
        else:
            j = self._filtration_dict[t]

        # Only add intervals of length 0 if
        # strict mode is not enabled.
        if i != j or (not self._strict):
            intervals[k].append((i, j))
            self.pairs.append((s, t))

    def _remove_pivot_rows(self, s, simplices):
        r"""
        Return the boundary chain of a simplex,
        from which pivot elements have been removed.

        This method implements the subroutine of the same name
        in [ZC2005]_. This method should not be called by users,
        it is used in the :meth:`_persistent_homology` method.

        TESTS::

            sage: l = [([0], 0), ([1], 0), ([2], 1), ([3], 1), ([0, 1], 1), ([1, 2], 1),
            ....:      ([0, 3], 2), ([2, 3], 2), ([0, 2], 3), ([0, 1, 2], 4)]
            sage: X = FilteredSimplicialComplex(l)
            sage: X._persistent_homology()                                              # needs sage.modules
            [[(0, 1), (1, 2), (0, +Infinity)], [(3, 4), (2, +Infinity)], []]
            sage: X._remove_pivot_rows(Simplex([0,1,2]), list(X._filtration_dict))      # needs sage.modules
            0
            sage: X.insert([0,2,3],5)
            sage: X._remove_pivot_rows(Simplex([0,2,3]), list(X._filtration_dict))      # needs sage.modules
            B[(2, 3)]
        """
        d = self._chaingroup()
        # Handle the case when the simplex is a vertex
        if s.dimension() == 0:
            return d

        # Initialize the boundary chain
        for i, f in enumerate(s.faces()):
            d += (-1)**i * self._chaingroup(f)

        # Remove all unmarked elements
        for s, x_s in d:
            j = self._index_of_simplex[s]
            if not self._marked[j]:
                d = d - x_s * self._chaingroup(s)

        # Reduce d until it is empty or until the simplex
        # with maximum index in the complex among all
        # nonzero terms is not in T.
        while d != 0:
            max_index = self._max_index(d)
            t = simplices[max_index]

            if not self._T[max_index]:
                break

            c = self._T[max_index][1]
            q = c[t]
            d = d - ((q**(-1)) * c)

        return d

    def _max_index(self, d):
        r"""
        Return the maximal index of all simplices with nonzero
        coefficient in ``d``.

        This method is called in :meth:`_remove_pivot_rows` and
        :meth:`_persistent_homology`. It should not be called by users
        outside of those methods.

        TESTS::

            sage: # needs sage.modules
            sage: X = FilteredSimplicialComplex([([0], 0), ([1], 5), ([0, 1], 18), ([0, 2, 3], 32)])
            sage: X._persistent_homology()
            [[(5, 18), (0, +Infinity)], [], []]
            sage: a = X._chaingroup(Simplex([0, 1]))
            sage: b = X._chaingroup(Simplex([0, 3]))
            sage: d = a + b
            sage: X._max_index(d)
            6
        """
        currmax = -1
        for s, x_s in d:
            j = self._index_of_simplex[s]
            currmax = max(j, currmax)
        return currmax

    def persistence_intervals(self, dimension, field=2, strict=True, verbose=None):
        r"""
        Return the list of `d`-dimensional homology elements.

        INPUT:

        - ``dimension`` -- integer; dimension `d` for which to
          return intervals
        - ``field`` -- prime number (default: 2); modulo which persistent
          homology is computed
        - ``strict`` -- boolean (default: ``True``); if ``False``, takes into account
          intervals of persistence 0
        - ``verbose`` -- (optional) if ``True``, print the steps of the
          persistent homology computation; the default is the verbosity
          of ``self``

        EXAMPLES::

            sage: X = FilteredSimplicialComplex([([0], 0), ([1], 1), ([0,1], 2)])
            sage: X.persistence_intervals(0)                                            # needs sage.modules
            [(1, 2), (0, +Infinity)]
        """
        if verbose is None:
            verbose = self._verbose
        intervals = self._persistent_homology(field, strict, verbose=verbose)
        if dimension < len(intervals):
            return intervals[dimension][:]
        else:
            return []

    def betti_number(self, k, a, b, field=2, strict=True, verbose=None):
        r"""
        Return the ``k``-dimensional Betti number from ``a`` to ``a + b``.

        INPUT:

        - ``k`` -- the dimension for the Betti number
        - ``a`` -- the lower filtration value
        - ``b`` -- the size of the interval
        - ``field`` -- prime number (default: 2); modulo which persistent
          homology is computed
        - ``strict`` -- boolean (default: ``True``); if ``False``, takes into account
          intervals of persistence 0
        - ``verbose`` -- (optional) if ``True``, print the steps of the
          persistent homology computation; the default is the verbosity
          of ``self``

        The Betti number `\beta_k^{a,a+b}` counts the number of
        homology elements which are alive throughout the whole
        duration ``[a, a+b]``.

        EXAMPLES::

            sage: X = FilteredSimplicialComplex([([0], 0), ([1], 0), ([0,1], 2)])
            sage: X.betti_number(0, 0.5, 1)                                             # needs sage.modules
            2
            sage: X.betti_number(0, 1.5, 1)                                             # needs sage.modules
            1

        If an element vanishes at time ``a + b`` exactly,
        it does not count towards the Betti number::

            sage: X = FilteredSimplicialComplex([([0], 0), ([1], 0), ([0,1], 2)])
            sage: X.betti_number(0, 1.5, 0.5)                                           # needs sage.modules
            1
        """
        if verbose is None:
            verbose = self._verbose
        intervals = self._persistent_homology(field, strict, verbose=verbose)
        return Integer(sum(1 for i, j in intervals[k]
                           if (i <= a and a + b < j) and a >= 0))

    def _repr_(self):
        """
        Print representation.

        If there are more than 10 simplices or vertices, only prints the
        count for each.

        EXAMPLES::

            sage: X = FilteredSimplicialComplex([([0], 0), ([1], 0), ([0, 1], 1)])
            sage: X
            Filtered complex on vertex set (0, 1) and with simplices ((0,) : 0), ((1,) : 0), ((0, 1) : 1)
            sage: X.insert([0, 1, 2, 3, 4], 8)
            sage: X
            Filtered complex on 5 vertices and with 31 simplices
        """
        vert_count = len(self._vertices)
        simp_count = len(self._filtration_dict)
        if simp_count > 10 or vert_count > 10:
            vertex_string = "on {} vertices".format(vert_count)
            simplex_string = "with {} simplices".format(simp_count)
        else:
            vertex_string = "on vertex set {}".format(tuple(sorted(self._vertices)))
            simplex_string = "with simplices "
            simplex_list = ["({} : {})".format(s, self._filtration_dict[s]) for s in self._filtration_dict]
            simplex_string += ", ".join(simplex_list)

        return "Filtered complex " + vertex_string + " and " + simplex_string

    def _simplicial_(self):
        """
        Return the associated simplicial complex.

        All simplices of the filtered simplicial complex are
        included in the resulting simplicial complex.

        EXAMPLES::

            sage: l = [([0],0), ([1],0), ([2],1), ([3],1), ([0, 1],1), ([1, 2],1), ([0, 3],2),
            ....:      ([2, 3],2), ([0, 2],3), ([0, 1, 2],4), ([0, 2, 3],5)]
            sage: a = FilteredSimplicialComplex(l)
            sage: b = SimplicialComplex(a)
            sage: b
            Simplicial complex with vertex set (0, 1, 2, 3) and facets {(0, 1, 2), (0, 2, 3)}
        """
        return SimplicialComplex(self._filtration_dict)



# sage_setup: distribution = sagemath-graphs
# distutils: language = c++
# distutils: extra_compile_args = -std=c++11

#*****************************************************************************
#       Copyright (C) 2015 Michele Borassi michele.borassi@imtlucca.it
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************
from libcpp.vector cimport vector
from libcpp.pair cimport pair

cdef extern from "boost/graph/adjacency_list.hpp" namespace "boost":
    cdef cppclass vecS:
        pass
    cdef cppclass listS:
        pass
    cdef cppclass slistS:
        pass
    cdef cppclass setS:
        pass
    cdef cppclass multisetS:
        pass
    cdef cppclass hash_setS:
        pass
    cdef cppclass undirectedS:
        pass
    cdef cppclass directedS:
        pass
    cdef cppclass bidirectionalS:
        pass
    cdef cppclass no_property:
        pass
    cdef cppclass edge_weight_t:
        pass
    cdef cppclass property[T, U]:
        pass


cdef extern from "boost_interface.cpp":
    ctypedef unsigned long v_index
    ctypedef unsigned long e_index

    cdef cppclass result_ec:
        v_index ec
        vector[v_index] edges

    cdef cppclass result_cc:
        float average_clustering_coefficient
        vector[float] clust_of_v

    cdef cppclass result_distances:
        vector[double] distances
        vector[v_index] predecessors

    cdef cppclass BoostGraph[OutEdgeListS, VertexListS, DirectedS, EdgeListS, EdgeProperty]:
        BoostGraph()
        void add_vertex()
        v_index num_verts()
        void add_edge(v_index u, v_index v)
        void add_edge(v_index u, v_index v, double w)
        vector[pair[v_index, pair[v_index, double]]] edge_list()
        e_index num_edges()
        result_ec edge_connectivity()
        double clustering_coeff(v_index v)
        result_cc clustering_coeff_all()
        vector[v_index] dominator_tree(v_index v)
        vector[v_index] bandwidth_ordering(bool)
        vector[v_index] kruskal_min_spanning_tree()
        vector[v_index] prim_min_spanning_tree()
        vector[vector[v_index]] blocks_and_cut_vertices()
        result_distances dijkstra_shortest_paths(v_index s)
        result_distances bellman_ford_shortest_paths(v_index s)
        vector[vector[double]] johnson_shortest_paths()
        vector[vector[double]] floyd_warshall_shortest_paths()

ctypedef property[edge_weight_t, double] EdgeWeight

ctypedef BoostGraph[vecS, vecS, undirectedS,    vecS, no_property] BoostVecGraph
ctypedef BoostGraph[vecS, vecS, bidirectionalS, vecS, no_property] BoostVecDiGraph

ctypedef BoostGraph[vecS, vecS, undirectedS,    vecS, EdgeWeight] BoostVecWeightedGraph
ctypedef BoostGraph[vecS, vecS, directedS,      vecS, EdgeWeight] BoostVecWeightedDiGraphU
ctypedef BoostGraph[vecS, vecS, bidirectionalS, vecS, EdgeWeight] BoostVecWeightedDiGraph

ctypedef BoostGraph[setS, vecS, undirectedS,    vecS, no_property] BoostSetGraph

ctypedef fused BoostVecGenGraph:
    BoostVecGraph
    BoostVecDiGraph

ctypedef fused BoostWeightedGraph:
    BoostVecWeightedGraph
    BoostVecWeightedDiGraph
    BoostVecWeightedDiGraphU

ctypedef fused BoostGenGraph:
    BoostVecGraph
    BoostVecDiGraph
    BoostGraph[vecS, vecS, directedS,      vecS, no_property]
    BoostSetGraph
    BoostGraph[setS, vecS, directedS,      vecS, no_property]
    BoostGraph[setS, vecS, bidirectionalS, vecS, no_property]



# sage_setup: distribution = sagemath-graphs
# sage.doctest: needs sage.graphs sage.groups
r"""
Knots

AUTHORS:

- Miguel Angel Marco Buzunariz
- Amit Jamadagni
"""

# ****************************************************************************
#       Copyright (C) 2014   Travis Scrimshaw <tscrim at ucdavis.edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.knots.link import Link
from sage.knots.knot_table import small_knots_table
from sage.knots.gauss_code import (recover_orientations, dowker_to_gauss,
                                   rectangular_diagram)

from sage.structure.parent import Parent
from sage.structure.element import Element
from sage.misc.fast_methods import Singleton
from sage.misc.inherit_comparison import InheritComparisonClasscallMetaclass
from sage.categories.monoids import Monoids

# We need Link to be first in the MRO in order to use its equality, hash, etc.


class Knot(Link, Element, metaclass=InheritComparisonClasscallMetaclass):
    r"""
    A knot.

    A knot is defined as embedding of the circle `\mathbb{S}^1` in the
    3-dimensional sphere `\mathbb{S}^3`, considered up to ambient isotopy.
    They represent the physical idea of a knotted rope, but with the
    particularity that the rope is closed. That is, the ends of the rope
    are joined.

    .. SEEALSO::

        :class:`Link`

    INPUT:

    - ``data`` -- see :class:`Link` for the allowable inputs
    - ``check`` -- boolean (default: ``True``); if ``True``, make sure
      that the data define a knot, not a link

    EXAMPLES:

    We construct the knot `8_{14}` and compute some invariants::

        sage: B = BraidGroup(4)
        sage: K = Knot(B([1,1,1,2,-1,2,-3,2,-3]))

    .. PLOT::
        :width: 300 px

        B = BraidGroup(4)
        K = Knot(B([1,1,1,2,-1,2,-3,2,-3]))
        sphinx_plot(K.plot())

    ::

        sage: K.alexander_polynomial()
        -2*t^-2 + 8*t^-1 - 11 + 8*t - 2*t^2
        sage: K.jones_polynomial()                                                      # needs sage.symbolic
        t^7 - 3*t^6 + 4*t^5 - 5*t^4 + 6*t^3 - 5*t^2 + 4*t + 1/t - 2
        sage: K.determinant()
        31
        sage: K.signature()
        -2
        sage: K.colored_jones_polynomial(2)  # long time
        q^-1 - 2 + 4*q - 5*q^2 + 6*q^3 - 5*q^4 + 4*q^5 - 3*q^6 + q^7

    REFERENCES:

    - :wikipedia:`Knot_(mathematics)`
    """
    @staticmethod
    def __classcall_private__(self, data, check=True):
        """
        Make sure this is an instance of the element class
        of :class:`Knots`.

        EXAMPLES::

            sage: B = BraidGroup(8)
            sage: K = Knot(B([-1, -1, -1, 2, 1, -2, 3, -2, 3]))
            sage: type(K)
            <class 'sage.knots.knot.Knots_with_category.element_class'>
        """
        return Knots().element_class(data, check=check)

    def __init__(self, data, check=True):
        """
        Initialize ``self``.

        TESTS::

            sage: B = BraidGroup(8)
            sage: K = Knot(B([1, -2, 1, -2]))
            sage: TestSuite(K).run()
            sage: K = Knot([[1, 1, 2, 2]])
            sage: TestSuite(K).run()

        The following is not a knot: it has two components. ::

            sage: Knot([[[1, 2], [-2, -1]], [1, -1]])
            Traceback (most recent call last):
            ...
            ValueError: the input has more than 1 connected component

            sage: Knot([[[1, 2], [-2, -1]], [1, -1]], check=False)
            Knot represented by 2 crossings
        """
        Element.__init__(self, Knots())
        Link.__init__(self, data)
        if check:
            if self.number_of_components() != 1:
                raise ValueError("the input has more than 1 connected "
                                 "component")

    def _repr_(self):
        """
        Return a string representation.

        EXAMPLES::

            sage: B = BraidGroup(8)
            sage: K = Knot(B([1, 2, 1, 2]))
            sage: K
            Knot represented by 4 crossings
            sage: K = Knot([[1, 7, 2, 6], [7, 3, 8, 2], [3, 11, 4, 10],
            ....:           [11, 5, 12, 4], [14, 5, 1, 6], [13, 9, 14, 8],
            ....:           [12, 9, 13, 10]])
            sage: K
            Knot represented by 7 crossings
        """
        pd_len = len(self.pd_code())
        return 'Knot represented by {} crossings'.format(pd_len)

    def _unicode_art_(self):
        """
        Return unicode art for the knot.

        EXAMPLES::

            sage: W = Knots()
            sage: K = W.from_dowker_code([-4,-6,-2])
            sage: unicode_art(K)
             ╭─╮
            ╭──│╮
            │╰╮││
            │ ╰─╯
            ╰──╯

            sage: G = [-1, 2, -3, 4, 6, -7, 8, 1, -2, 3, -4, -5, 7, -8, 5, -6]
            sage: K = Knots().from_gauss_code(G)
            sage: unicode_art(K)
               ╭─────╮
               │╭─────╮
              ╭─│╮   ││
             ╭─╯││   ││
             │╰─╯│   ││
             │   ╰──╮││
             │     ╭│╯│
             │    ╭│╯ │
            ╭─────│╯  │
            │╰────╯   │
            ╰─────────╯

        TESTS::

            sage: W = Knots()
            sage: unicode_art(W.one())
            ╭╮
            ╰╯
        """
        style = 2  # among 0, 1, 2 (how to display crossings, see below)
        gauss = self.gauss_code()
        if not gauss:
            gauss = []
        else:
            gauss = gauss[0]

        graphe, (hori, vert) = rectangular_diagram(gauss)
        maxx, maxy = 0, 0
        for a, b in graphe:
            maxx = max(a, maxx)
            maxy = max(b, maxy)
        M = [[" " for a in range(maxy + 1)] for b in range(maxx + 1)]
        for a, b in graphe:
            (x, y), (xx, yy) = graphe.neighbors((a, b))
            if x != a:
                x, y, xx, yy = xx, yy, x, y
            if y < b:
                if xx < a:
                    M[a][b] = "╯"
                else:
                    M[a][b] = "╮"
            else:
                if xx < a:
                    M[a][b] = "╰"
                else:
                    M[a][b] = "╭"

        for ab, cd in graphe.edge_iterator(labels=False):
            a, b = ab
            c, d = cd
            if a == c:
                b, d = sorted((b, d))
                for i in range(b + 1, d):
                    M[a][i] = "─"
            else:
                a, c = sorted((a, c))
                for i in range(a + 1, c):
                    M[i][b] = "│"

        if style == 0:
            H = "┿"
            V = "╂"
        elif style == 1:
            H = "━"
            V = "┃"
        elif style == 2:
            H = "─"
            V = "│"

        for x, y in hori:
            M[x][y] = H
        for x, y in vert:
            M[x][y] = V

        from sage.typeset.unicode_art import UnicodeArt
        return UnicodeArt([''.join(ligne) for ligne in M])

    def dt_code(self):
        """
        Return the DT code of ``self``.

        ALGORITHM:

        The DT code is generated by the following way:

        Start moving along the knot, as we encounter the crossings we
        start numbering them, so every crossing has two numbers assigned to
        it once we have traced the entire knot. Now we take the even number
        associated with every crossing.

        The following sign convention is to be followed:

        Take the even number with a negative sign if it is an overcrossing
        that we are encountering.

        OUTPUT: DT code representation of the knot

        EXAMPLES::

            sage: K = Knot([[1,5,2,4],[5,3,6,2],[3,1,4,6]])
            sage: K.dt_code()
            [4, 6, 2]

            sage: B = BraidGroup(4)
            sage: K = Knot(B([1, 2, 1, 2]))
            sage: K.dt_code()
            [4, -6, 8, -2]

            sage: K = Knot([[[1, -2, 3, -4, 5, -1, 2, -3, 4, -5]],
            ....:          [1, 1, 1, 1, 1]])
            sage: K.dt_code()
            [6, 8, 10, 2, 4]
        """
        b = self.braid().Tietze()
        N = len(b)
        label = [0 for i in range(2 * N)]
        string = 1
        next_label = 1
        type1 = 0
        crossing = 0
        while next_label <= 2 * N:
            string_found = False
            for i in range(crossing, N):
                if abs(b[i]) == string or abs(b[i]) == string - 1:
                    string_found = True
                    crossing = i
                    break
            if not string_found:
                for i in range(crossing):
                    if abs(b[i]) == string or abs(b[i]) == string - 1:
                        string_found = True
                        crossing = i
                        break
            assert label[2 * crossing + next_label % 2] != 1, "invalid knot"

            label[2 * crossing + next_label % 2] = next_label
            next_label += 1
            if type1 == 0:
                if b[crossing] < 0:
                    type1 = 1
                else:
                    type1 = -1
            else:
                type1 = -type1
                if (abs(b[crossing]) == string) ^ (b[crossing] * type1 < 0):
                    if next_label % 2:
                        label[2 * crossing] = label[2 * crossing] * -1
            if abs(b[crossing]) == string:
                string += 1
            else:
                string -= 1
            crossing += 1
        code = [0 for _ in range(N)]
        for i in range(N):
            for j in range(N):
                if label[2 * j + 1] == 2 * i + 1:
                    code[i] = label[2 * j]
                    break
        return code

    def arf_invariant(self):
        """
        Return the Arf invariant.

        EXAMPLES::

            sage: B = BraidGroup(4)
            sage: K = Knot(B([-1, 2, 1, 2]))
            sage: K.arf_invariant()
            0
            sage: B = BraidGroup(8)
            sage: K = Knot(B([-2, 3, 1, 2, 1, 4]))
            sage: K.arf_invariant()
            0
            sage: K = Knot(B([1, 2, 1, 2]))
            sage: K.arf_invariant()
            1
        """
        a = self.alexander_polynomial()(-1)
        if (a % 8) == 1 or (a % 8) == 7:
            return 0
        return 1

    def colored_jones_polynomial(self, N, variab=None, try_inverse=True):
        r"""
        Return the colored Jones polynomial of the trace closure of the braid.

        INPUT:

        - ``N`` -- integer; the number of colors
        - ``variab`` -- (default: `q`) the variable in the resulting
          Laurent polynomial
        - ``try_inverse`` -- boolean (default: ``True``); if ``True``,
          attempt a faster calculation by using the inverse of the braid

        ALGORITHM:

        The algorithm used is described in [HL2018]_ for the corresponding
        braid representation. We follow their notation, but work in a
        suitable free algebra over a Laurent polynomial ring in one
        variable to simplify bookkeeping.

        EXAMPLES::

            sage: W = Knots()
            sage: K = W.from_dowker_code([-4,-6,-2])
            sage: K.colored_jones_polynomial(2)
            -q^-4 + q^-3 + q^-1
            sage: K.colored_jones_polynomial(2, 't')
            -t^-4 + t^-3 + t^-1
            sage: R.<t> = LaurentPolynomialRing(ZZ)
            sage: K.colored_jones_polynomial(2, -t)
            -t^-4 - t^-3 - t^-1

            sage: R.<t> = ZZ[]
            sage: K.colored_jones_polynomial(2, t+1)
            (t^3 + 3*t^2 + 4*t + 1)/(t^4 + 4*t^3 + 6*t^2 + 4*t + 1)
        """
        return self.braid().colored_jones_polynomial(N=N, variab=variab,
                                                     try_inverse=try_inverse)

    def connected_sum(self, other):
        r"""
        Return the oriented connected sum of ``self`` and ``other``.

        .. NOTE::

            We give the knots an orientation based upon the braid
            representation.

        INPUT:

        - ``other`` -- a knot

        OUTPUT: a knot equivalent to the connected sum of ``self`` and ``other``

        EXAMPLES::

            sage: B = BraidGroup(2)
            sage: trefoil = Knot(B([1,1,1]))
            sage: K = trefoil.connected_sum(trefoil); K
            Knot represented by 6 crossings
            sage: K.braid()
            s0^3*s1^-1*s0^3*s1

        .. PLOT::
            :width: 300 px

            B = BraidGroup(2)
            trefoil = Knot(B([1,1,1]))
            K = trefoil.connected_sum(trefoil)
            sphinx_plot(K.plot())

        ::

            sage: rev_trefoil = Knot(B([-1,-1,-1]))
            sage: K2 = trefoil.connected_sum(rev_trefoil); K2
            Knot represented by 6 crossings
            sage: K2.braid()
            s0^3*s1^-1*s0^-3*s1

        .. PLOT::
            :width: 300 px

            B = BraidGroup(2)
            t = Knot(B([1,1,1]))
            tr = Knot(B([-1,-1,-1]))
            K2 = t.connected_sum(tr)
            sphinx_plot(K2.plot())

        Observe that both knots have according ``dowker_notation`` (showing that
        the constructing from DT-code may not be unique for non prime knots, see
        :meth:`from_dowker_code`)::

            sage: K.dowker_notation()
            [(4, 1), (2, 5), (6, 3), (10, 7), (8, 11), (12, 9)]
            sage: K2.dowker_notation()
            [(4, 1), (2, 5), (6, 3), (7, 10), (11, 8), (9, 12)]
            sage: K.homfly_polynomial() == K2.homfly_polynomial()                       # needs libhomfly
            False

        TESTS::

            sage: B = BraidGroup(2)
            sage: trivial = Knots().one()
            sage: trivial * trivial
            Knot represented by 0 crossings
            sage: trefoil = Knot(B([1,1,1]))
            sage: trefoil * trivial
            Knot represented by 3 crossings
            sage: trefoil * trefoil
            Knot represented by 6 crossings

        REFERENCES:

        - :wikipedia:`Connected_sum`
        """
        from sage.functions.generalized import sign
        ogc1 = self.oriented_gauss_code()
        ogc2 = other.oriented_gauss_code()
        if not ogc1[0]:
            return other
        if not ogc2[0]:
            return self
        # how much we have to "displace" the numbering of the
        # crossings of other
        m1 = max(abs(i) for i in ogc1[0][0])
        m2 = min(abs(i) for i in ogc2[0][0])
        n = m1 - m2 + 1
        # construct the oriented Gauss code of the result
        ogc2_0_0 = [a + int(sign(a)) * n for a in ogc2[0][0]]
        nogc = [[ogc1[0][0] + ogc2_0_0], ogc1[1] + ogc2[1]]
        return type(self)(nogc)

    _mul_ = connected_sum


class Knots(Singleton, Parent):
    """
    The set for all knots, as a monoid for the connected sum.
    """
    def __init__(self):
        """
        TESTS::

            sage: S = Knots()
            sage: S.cardinality()
            +Infinity
            sage: TestSuite(S).run()
        """
        Parent.__init__(self, category=Monoids().Infinite())

    def _repr_(self):
        r"""
        TESTS::

            sage: Knots()
            Knots
        """
        return "Knots"

    def one(self):
        """
        Return the unit of the monoid.

        This is the trivial knot.

        EXAMPLES::

            sage: Knots().one()
            Knot represented by 0 crossings
        """
        return self.element_class([])

    def an_element(self):
        """
        Return the trefoil knot.

        EXAMPLES::

            sage: Knots().an_element()
            Knot represented by 3 crossings
        """
        return self.element_class([[1, 5, 2, 4], [5, 3, 6, 2], [3, 1, 4, 6]])

    def from_gauss_code(self, gauss):
        """
        Build a knot from a signed Gauss code.

        This makes some arbitrary choice of orientation.

        INPUT:

        - ``gauss`` -- a signed Gauss code

        OUTPUT: a knot

        EXAMPLES::

            sage: W = Knots()
            sage: K1 = W.from_gauss_code([2, -1, 3, -2, 1, -3])
            sage: K1.alexander_polynomial()
            t^-1 - 1 + t
        """
        orientations = recover_orientations(gauss)[3]
        return Knot([[gauss], orientations])

    def from_dowker_code(self, code):
        """
        Build a knot from a Dowker-Thistlethwaite code.

        The Dowker-Thistlethwaite code of a knot diagram is defined as follows.

        Start following the knot diagram at some regular point. Label the
        crossings by a number (starting from number 1) in the order in
        which they are met. At the end, every crossing gets numbered
        twice, once by an even number and once by an odd number. When
        meeting an over-crossing with even number, use instead the
        negative of this even number as label.

        Then the set of crossings gives a set of pairs (odd,
        even). Sort this set according to the odd component, and then
        keep only the even components in the same order. This is the
        Dowker-Thistlethwaite code.

        INPUT:

        - ``code`` -- list of signed even numbers; the Dowker-Thistlethwaite
          code of a knot

        OUTPUT: a knot

        .. WARNING::

            In general the Dowker-Thistlethwaite code does not describe a knot
            uniquely. It is not only insensitive on mirror images, but may also
            mix up non prime knots. For example ``[4, 6, 2, 10, 12, 8]`` describes
            the connected sum of two trefoil knots, as well as the connected sum
            of a trefoil with its mirror (see the corresponding example in the
            documentation of :meth:`connected_sum`).

        EXAMPLES::

            sage: W = Knots()
            sage: K1 = W.from_dowker_code([8,10,2,12,4,6])
            sage: K1.dowker_notation()
            [(5, 2), (9, 4), (11, 6), (1, 8), (3, 10), (7, 12)]

            sage: W.from_dowker_code([6,10,14,12,16,2,18,4,8])
            Knot represented by 9 crossings

            sage: W.from_dowker_code([4,8,10,-14,2,-16,-18,-6,-12])
            Knot represented by 9 crossings

            sage: K3 = W.from_dowker_code([6,-12,2,8,-4,-10]); K3
            Knot represented by 6 crossings
            sage: K3.dowker_notation()
            [(5, 2), (4, 9), (1, 6), (7, 8), (10, 11), (12, 3)]

        .. SEEALSO:: :meth:`~sage.knots.knot.Knot.dowker_notation`

        REFERENCES:

        - :wikipedia:`Dowker_notation`

        - http://katlas.org/wiki/DT_(Dowker-Thistlethwaite)_Codes
        """
        gauss = dowker_to_gauss(code)
        orientations = recover_orientations(gauss)[3]
        return Knot([[gauss], orientations])

    def from_table(self, n, k):
        """
        Return a knot from its index in the Rolfsen table.

        INPUT:

        - ``n`` -- the crossing number
        - ``k`` -- positive integer

        OUTPUT: the knot `K_{n,k}` in the Rolfsen table

        EXAMPLES::

            sage: K1 = Knots().from_table(6,3); K1
            Knot represented by 6 crossings
            sage: K1.alexander_polynomial()
            t^-2 - 3*t^-1 + 5 - 3*t + t^2

            sage: K2 = Knots().from_table(8,4); K2
            Knot represented by 9 crossings
            sage: K2.determinant()
            19
            sage: K2.signature()
            2

            sage: K3 = Knots().from_table(10,56); K3
            Knot represented by 11 crossings
            sage: K3.jones_polynomial()                                                 # needs sage.symbolic
            t^10 - 3*t^9 + 6*t^8 - 9*t^7 + 10*t^6 - 11*t^5 + 10*t^4 - 7*t^3
            + 5*t^2 - 2*t + 1

            sage: K4 = Knots().from_table(10,100)
            sage: K4.genus()
            4

        TESTS::

            sage: Knots().from_table(6,6)
            Traceback (most recent call last):
            ...
            ValueError: not found in the knot table

            sage: Knots().from_table(12,6)
            Traceback (most recent call last):
            ...
            ValueError: more than 10 crossings, not in the knot table

        REFERENCES:

        - KnotAtlas, http://katlas.math.toronto.edu/wiki/The_Rolfsen_Knot_Table
        """
        if n > 10:
            raise ValueError('more than 10 crossings, not in the knot table')
        if (n, k) in small_knots_table:
            m, word = small_knots_table[(n, k)]

            from sage.groups.braid import BraidGroup

            G = BraidGroup(m)
            return Knot(G(word))
        else:
            raise ValueError('not found in the knot table')

    Element = Knot
